'use strict'

var forEach = require('transduce/array/forEach')

module.exports = function(_r){
  // Array Functions
  // ---------------
  _r.mixin({
    forEach: forEach,
    each: forEach,
    find: find,
    detect: find,
    every: every,
    all: every,
    some: some,
    any: some,
    contains: contains,
    include: contains,
    findWhere: findWhere,
    push: require('transduce/array/push'),
    unshift: require('transduce/array/unshift'),
    at: at,
    slice: require('transduce/array/slice'),
    initial: require('transduce/array/initial'),
    last: last
  })

  var iteratee = _r.iteratee,
      resolveSingleValue = _r.resolveSingleValue,
      _ = _r._

  // Return the first value which passes a truth test. Aliased as `detect`.
  var _find = require('transduce/array/find')
  function find(predicate) {
     /*jshint validthis:true*/
     resolveSingleValue(this)
     return _find(iteratee(predicate))
  }

  // Determine whether all of the elements match a truth test.
  // Aliased as `all`.
  var _every = require('transduce/array/every')
  function every(predicate) {
     /*jshint validthis:true*/
    resolveSingleValue(this)
    return _every(iteratee(predicate))
  }

  // Determine if at least one element in the object matches a truth test.
  // Aliased as `any`.
  var _some = require('transduce/array/some')
  function some(predicate) {
     /*jshint validthis:true*/
    resolveSingleValue(this)
    return _some(iteratee(predicate))
  }

  // Determine if contains a given value (using `===`).
  // Aliased as `include`.
  function contains(target) {
     /*jshint validthis:true*/
    return some.call(this, function(x){ return x === target })
  }

  // Convenience version of a common use case of `find`: getting the first object
  // containing specific `key:value` pairs.
  function findWhere(attrs) {
     /*jshint validthis:true*/
    return find.call(this, _.matches(attrs))
  }

  // Retrieves the value at the given index. Resolves as single value.
  var _slice = require('transduce/array/slice')
  function at(idx){
     /*jshint validthis:true*/
    resolveSingleValue(this)
    return _slice(idx, idx+1)
  }

  // Get the last element. Passing **n** will return the last N  values.
  // Note that no items will be sent until completion.
  var _last = require('transduce/array/last')
  function last(n) {
    if(n === void 0){
     /*jshint validthis:true*/
      resolveSingleValue(this)
    }
    return _last(n)
  }
}
